@@ -31,7 +31,7 @@
 #include <gnutls_errors.h>
 #include <aes-x86.h>
 #include <sha-x86.h>
-#include <x86.h>
+#include <x86-common.h>
 
 struct aes_ctx {
 	AES_KEY expanded_key;
@@ -39,8 +39,6 @@ struct aes_ctx {
 	int enc;
 };
 
-unsigned int _gnutls_x86_cpuid_s[4];
-
 static int
 aes_cipher_init(gnutls_cipher_algorithm_t algorithm, void **_ctx, int enc)
 {
@@ -117,6 +115,7 @@ static void aes_deinit(void *_ctx)
 {
 	struct aes_ctx *ctx = _ctx;
 	
+	zeroize_temp_key(ctx, sizeof(*ctx));
 	gnutls_free(ctx);
 }
 
