@@ -2597,6 +2597,11 @@ int gnutls_handshake(gnutls_session_t session)
 void
 gnutls_handshake_set_timeout(gnutls_session_t session, unsigned int ms)
 {
+	if (IS_DTLS(session)) {
+		gnutls_dtls_set_timeouts(session, 1000, ms);
+		return;
+	}
+
 	if (ms == GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT)
 		ms = 40 * 1000;
 	session->internals.handshake_timeout_ms = ms;
