@@ -39,13 +39,60 @@
 #include <gnutls_pk.h>
 #include <stdbool.h>
 
-/* Checks if two certs are identical.  Return 1 on match. */
+/* Checks if two certs have the same name and the same key.  Return 1 on match. 
+ * If @is_ca is zero then this function is identical to _gnutls_check_if_same_cert()
+ */
+bool
+_gnutls_check_if_same_key(gnutls_x509_crt_t cert1,
+			  gnutls_x509_crt_t cert2,
+			  unsigned is_ca)
+{
+	int ret;
+	bool result;
+
+	if (is_ca == 0)
+		return _gnutls_check_if_same_cert(cert1, cert2);
+
+	ret = _gnutls_is_same_dn(cert1, cert2);
+	if (ret == 0)
+		return 0;
+
+	if (cert1->raw_spki.size > 0 && (cert1->raw_spki.size == cert2->raw_spki.size) &&
+	    (memcmp(cert1->raw_spki.data, cert2->raw_spki.data, cert1->raw_spki.size) == 0))
+		result = 1;
+	else
+		result = 0;
+
+	return result;
+}
+
+bool
+_gnutls_check_if_same_key2(gnutls_x509_crt_t cert1,
+			   gnutls_datum_t * cert2bin)
+{
+	int ret;
+	gnutls_x509_crt_t cert2;
+
+	ret = gnutls_x509_crt_init(&cert2);
+	if (ret < 0)
+		return gnutls_assert_val(0);
+
+	ret = gnutls_x509_crt_import(cert2, cert2bin, GNUTLS_X509_FMT_DER);
+	if (ret < 0) {
+		gnutls_x509_crt_deinit(cert2);
+		return gnutls_assert_val(0);
+	}
+
+	ret = _gnutls_check_if_same_key(cert1, cert2, 1);
+
+	gnutls_x509_crt_deinit(cert2);
+	return ret;
+}
+
 bool
 _gnutls_check_if_same_cert(gnutls_x509_crt_t cert1,
 			   gnutls_x509_crt_t cert2)
 {
-	gnutls_datum_t cert1bin = { NULL, 0 }, cert2bin = {
-	NULL, 0};
 	int ret;
 	bool result;
 
@@ -53,29 +100,12 @@ _gnutls_check_if_same_cert(gnutls_x509_crt_t cert1,
 	if (ret == 0)
 		return 0;
 
-	ret = _gnutls_x509_der_encode(cert1->cert, "", &cert1bin, 0);
-	if (ret < 0) {
-		gnutls_assert();
-		result = 0;
-		goto cleanup;
-	}
-
-	ret = _gnutls_x509_der_encode(cert2->cert, "", &cert2bin, 0);
-	if (ret < 0) {
-		gnutls_assert();
-		result = 0;
-		goto cleanup;
-	}
-
-	if ((cert1bin.size == cert2bin.size) &&
-	    (memcmp(cert1bin.data, cert2bin.data, cert1bin.size) == 0))
+	if ((cert1->der.size == cert2->der.size) &&
+	    (memcmp(cert1->der.data, cert2->der.data, cert1->der.size) == 0))
 		result = 1;
 	else
 		result = 0;
 
-      cleanup:
-	_gnutls_free_datum(&cert1bin);
-	_gnutls_free_datum(&cert2bin);
 	return result;
 }
 
@@ -83,25 +113,14 @@ bool
 _gnutls_check_if_same_cert2(gnutls_x509_crt_t cert1,
 			    gnutls_datum_t * cert2bin)
 {
-	gnutls_datum_t cert1bin = { NULL, 0 };
-	int ret;
 	bool result;
 
-	ret = _gnutls_x509_der_encode(cert1->cert, "", &cert1bin, 0);
-	if (ret < 0) {
-		result = 0;
-		gnutls_assert();
-		goto cleanup;
-	}
-
-	if ((cert1bin.size == cert2bin->size) &&
-	    (memcmp(cert1bin.data, cert2bin->data, cert1bin.size) == 0))
+	if ((cert1->der.size == cert2bin->size) &&
+	    (memcmp(cert1->der.data, cert2bin->data, cert1->der.size) == 0))
 		result = 1;
 	else
 		result = 0;
 
-      cleanup:
-	_gnutls_free_datum(&cert1bin);
 	return result;
 }
 
@@ -130,7 +149,7 @@ check_if_ca(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer,
 	 */
 
 	ret =
-	    _gnutls_x509_get_signed_data(issuer->cert, "tbsCertificate",
+	    _gnutls_x509_get_signed_data(issuer->cert, &issuer->der, "tbsCertificate",
 					 &issuer_signed_data);
 	if (ret < 0) {
 		gnutls_assert();
@@ -138,7 +157,7 @@ check_if_ca(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer,
 	}
 
 	ret =
-	    _gnutls_x509_get_signed_data(cert->cert, "tbsCertificate",
+	    _gnutls_x509_get_signed_data(cert->cert, &cert->der, "tbsCertificate",
 					 &cert_signed_data);
 	if (ret < 0) {
 		gnutls_assert();
@@ -545,6 +564,7 @@ verify_crt(gnutls_x509_crt_t cert,
 			    gnutls_x509_crt_t * _issuer,
 			    time_t now,
 			    unsigned int *max_path,
+			    bool end_cert,
 			    gnutls_x509_name_constraints_t nc,
 			    gnutls_verify_output_function func)
 {
@@ -606,34 +626,37 @@ verify_crt(gnutls_x509_crt_t cert,
 			goto cleanup;
 		}
 
-		ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_DNSNAME, cert);
-		if (ret == 0) {
-			gnutls_assert();
-			goto nc_fail;
-		}
+		/* only check name constraints in server certificates, not CAs */
+		if (end_cert != 0) {
+			ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_DNSNAME, cert);
+			if (ret == 0) {
+				gnutls_assert();
+				goto nc_fail;
+			}
 
-		ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_RFC822NAME, cert);
-		if (ret == 0) {
-			gnutls_assert();
-			goto nc_fail;
-		}
+			ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_RFC822NAME, cert);
+			if (ret == 0) {
+				gnutls_assert();
+				goto nc_fail;
+			}
 
-		ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_DN, cert);
-		if (ret == 0) {
-			gnutls_assert();
-			goto nc_fail;
-		}
+			ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_DN, cert);
+			if (ret == 0) {
+				gnutls_assert();
+				goto nc_fail;
+			}
 
-		ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_URI, cert);
-		if (ret == 0) {
-			gnutls_assert();
-			goto nc_fail;
-		}
+			ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_URI, cert);
+			if (ret == 0) {
+				gnutls_assert();
+				goto nc_fail;
+			}
 
-		ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_IPADDRESS, cert);
-		if (ret == 0) {
-			gnutls_assert();
-			goto nc_fail;
+			ret = gnutls_x509_name_constraints_check_crt(nc, GNUTLS_SAN_IPADDRESS, cert);
+			if (ret == 0) {
+				gnutls_assert();
+				goto nc_fail;
+			}
 		}
 	}
 
@@ -675,7 +698,7 @@ verify_crt(gnutls_x509_crt_t cert,
 	}
 
 	ret =
-	    _gnutls_x509_get_signed_data(cert->cert, "tbsCertificate",
+	    _gnutls_x509_get_signed_data(cert->cert, &cert->der, "tbsCertificate",
 					 &cert_signed_data);
 	if (ret < 0) {
 		result = 0;
@@ -801,6 +824,9 @@ verify_crt(gnutls_x509_crt_t cert,
  * given issuer. It checks the DN fields and the authority
  * key identifier and subject key identifier fields match.
  *
+ * If the same certificate is provided at the @cert and @issuer fields,
+ * it will check whether the certificate is self-signed.
+ *
  * Returns: It will return true (1) if the given certificate is issued
  *   by the given issuer, and false (0) if not.  
  **/
@@ -864,8 +890,12 @@ _gnutls_verify_crt_status(const gnutls_x509_crt_t * certificate_list,
 		int j;
 
 		for (j = 0; j < tcas_size; j++) {
-			if (_gnutls_check_if_same_cert
-			    (certificate_list[i], trusted_cas[j]) != 0) {
+			/* we check for a certificate that may not be identical with the one
+			 * sent by the client, but will have the same name and key. That is
+			 * because it can happen that a CA certificate is upgraded from intermediate
+			 * CA to self-signed CA at some point. */
+			if (_gnutls_check_if_same_key
+			    (certificate_list[i], trusted_cas[j], i) != 0) {
 				/* explicit time check for trusted CA that we remove from
 				 * list. GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS
 				 */
@@ -913,10 +943,11 @@ _gnutls_verify_crt_status(const gnutls_x509_crt_t * certificate_list,
 	 */
 	output = 0;
 	max_path = MAX_VERIFY_DEPTH;
+
 	ret = verify_crt(certificate_list[clist_size - 1],
 					  trusted_cas, tcas_size, flags,
 					  &output, &issuer, now, &max_path,
-					  nc, func);
+					  clist_size==1?1:0, nc, func);
 	if (ret != 1) {
 		/* if the last certificate in the certificate
 		 * list is invalid, then the certificate is not
@@ -941,11 +972,12 @@ _gnutls_verify_crt_status(const gnutls_x509_crt_t * certificate_list,
 		if (!(flags & GNUTLS_VERIFY_ALLOW_ANY_X509_V1_CA_CRT)) {
 			flags |= GNUTLS_VERIFY_DO_NOT_ALLOW_X509_V1_CA_CRT;
 		}
+
 		if ((ret =
 		     verify_crt(certificate_list[i - 1],
 						 &certificate_list[i], 1,
 						 flags, &output, NULL, now,
-						 &max_path, nc, func)) != 1) {
+						 &max_path, i==1?1:0, nc, func)) != 1) {
 			gnutls_assert();
 			status |= output;
 			status |= GNUTLS_CERT_INVALID;
@@ -959,11 +991,47 @@ cleanup:
 	return status;
 }
 
+/* Returns true if the provided purpose is in accordance with the certificate.
+ */
+bool _gnutls_check_key_purpose(gnutls_x509_crt_t cert, const char *purpose)
+{
+	char oid[MAX_OID_SIZE];
+	size_t oid_size;
+	int ret;
+	unsigned i;
+
+	for (i=0;;i++) {
+		oid_size = sizeof(oid);
+		ret = gnutls_x509_crt_get_key_purpose_oid(cert, i, oid, &oid_size, NULL);
+		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+			if (i==0) {
+				/* no key purpose in certificate, assume ANY */
+				return 1;
+			} else {
+				gnutls_assert();
+				break;
+			}
+		} else if (ret < 0) {
+			gnutls_assert();
+			break;
+		}
+
+		if (strcmp(oid, purpose) == 0 || strcmp(oid, GNUTLS_KP_ANY) == 0) {
+			return 1;
+		}
+	}
+	return 0;
+}
+
 #ifdef ENABLE_PKCS11
 /* Verify X.509 certificate chain using a PKCS #11 token.
  *
  * Note that the return value is an OR of GNUTLS_CERT_* elements.
  *
+ * Unlike the non-PKCS#11 version, this function accepts a key purpose
+ * (from GNUTLS_KP_...). That is because in the p11-kit trust modules
+ * anchors are mixed and get assigned a purpose.
+ *
  * This function verifies a X.509 certificate list. The certificate
  * list should lead to a trusted certificate in order to be trusted.
  */
@@ -971,6 +1039,7 @@ unsigned int
 _gnutls_pkcs11_verify_crt_status(const char* url,
 				const gnutls_x509_crt_t * certificate_list,
 				unsigned clist_size,
+				const char *purpose,
 				unsigned int flags,
 				gnutls_verify_output_function func)
 {
@@ -1008,9 +1077,16 @@ _gnutls_pkcs11_verify_crt_status(const char* url,
 		i = 1;		/* do not replace the first one */
 
 	for (; i < clist_size; i++) {
-		if (gnutls_pkcs11_crt_is_known (url, certificate_list[i], 
-			GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE|
-			GNUTLS_PKCS11_OBJ_FLAG_COMPARE|GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED) != 0) {
+		unsigned vflags;
+
+		if (i == 0) /* in the end certificate do full comparison */
+			vflags = GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE|
+				GNUTLS_PKCS11_OBJ_FLAG_COMPARE|GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED;
+		else
+			vflags = GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE|
+				GNUTLS_PKCS11_OBJ_FLAG_COMPARE_KEY|GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED;
+
+		if (gnutls_pkcs11_crt_is_known (url, certificate_list[i], vflags) != 0) {
 
 			if (!(flags & GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS) &&
 				!(flags & GNUTLS_VERIFY_DISABLE_TIME_CHECKS)) {
@@ -1053,7 +1129,8 @@ _gnutls_pkcs11_verify_crt_status(const char* url,
 
 	/* check against issuer */
 	ret = gnutls_pkcs11_get_raw_issuer(url, certificate_list[clist_size - 1],
-		&raw_issuer, GNUTLS_X509_FMT_DER, GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE);
+			 		   &raw_issuer, GNUTLS_X509_FMT_DER,
+			 		   GNUTLS_PKCS11_OBJ_FLAG_OVERWRITE_TRUSTMOD_EXT|GNUTLS_PKCS11_OBJ_FLAG_PRESENT_IN_TRUSTED_MODULE);
 	if (ret < 0) {
 		gnutls_assert();
 		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE && clist_size > 2) {
@@ -1089,6 +1166,16 @@ _gnutls_pkcs11_verify_crt_status(const char* url,
 		goto cleanup;
 	}
 
+	if (purpose != NULL) {
+		ret = _gnutls_check_key_purpose(issuer, purpose);
+		if (ret != 1) {
+			gnutls_assert();
+			status |= GNUTLS_CERT_INVALID;
+			status |= GNUTLS_CERT_SIGNER_NOT_FOUND;
+			goto cleanup;
+		}
+	}
+
 	status = _gnutls_verify_crt_status(certificate_list, clist_size,
 				&issuer, 1, flags, func);
 
@@ -1360,10 +1447,12 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 	}
 
 	result =
-	    _gnutls_x509_get_signed_data(crl->crl, "tbsCertList",
+	    _gnutls_x509_get_signed_data(crl->crl, &crl->der, "tbsCertList",
 					 &crl_signed_data);
 	if (result < 0) {
 		gnutls_assert();
+		if (verify)
+			*verify |= GNUTLS_CERT_INVALID;
 		goto cleanup;
 	}
 
@@ -1372,6 +1461,8 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 				       &crl_signature);
 	if (result < 0) {
 		gnutls_assert();
+		if (verify)
+			*verify |= GNUTLS_CERT_INVALID;
 		goto cleanup;
 	}
 
@@ -1380,6 +1471,8 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 						 "signatureAlgorithm.algorithm");
 	if (result < 0) {
 		gnutls_assert();
+		if (verify)
+			*verify |= GNUTLS_CERT_INVALID;
 		goto cleanup;
 	}
 
@@ -1397,6 +1490,8 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 		result = 0;
 	} else if (result < 0) {
 		gnutls_assert();
+		if (verify)
+			*verify |= GNUTLS_CERT_INVALID;
 		goto cleanup;
 	}
 
@@ -1423,7 +1518,7 @@ gnutls_x509_crl_verify(gnutls_x509_crl_t crl,
 
 
       cleanup:
-	if (verify)
+	if (verify && *verify != 0)
 		*verify |= GNUTLS_CERT_INVALID;
 
 	_gnutls_free_datum(&crl_signed_data);
